import torch.nn as nn

__all__ = ['RRCNN_block', 'Recurrent_block']

class Recurrent_block(nn.Module):
    """
    Recurrent Block for CNN modules
    """
    def __init__(self, out_ch, t=2):
        super(Recurrent_block, self).__init__()

        self.t = t
        self.out_ch = out_ch
        self.conv = nn.Sequential(
            nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1, bias=True),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=False)
        )

    def forward(self, x):
        for i in range(self.t):
            if i == 0:
                x = self.conv(x)
            out = self.conv(x + x)
        return out


class RRCNN_block(nn.Module):
    """
    Recurrent Residual Convolutional Neural Network Block
    """
    def __init__(self, in_ch, out_ch, t=2):
        super(RRCNN_block, self).__init__()

        self.RCNN = nn.Sequential(
            Recurrent_block(out_ch, t=t),
            Recurrent_block(out_ch, t=t)
        )
        self.Conv = nn.Conv2d(in_ch, out_ch, kernel_size=1, stride=1, padding=0)

    def forward(self, x):
        x1 = self.Conv(x)
        x2 = self.RCNN(x1)
        out = x1 + x2
        return out


import torch.nn as nn
import torch
from torch.autograd import Variable
import torch.nn.functional as F
import numpy as np
from ..utils.dist import euclidean_dist

__all__ = ['ProtoNet']


class Flatten(nn.Module):
    def __init__(self):
        super(Flatten, self).__init__()

    def forward(self, x):
        return x.view(x.size(0), -1)

def load_protonet_conv(**kwargs):
    x_dim=kwargs['x_dim']
    y_dim=kwargs['y_dim']
    z_dim=kwargs['z_dim']

    def conv_block(in_ch,out_ch):
        return nn.Sequential(
                nn.Conv2d(in_ch,out_ch,1,stride=1),
                nn.BatchNorm2d(out_ch),
                nn.ReLU(),
                nn.MaxPool2d(2))
    encoder=nn.Sequential(
            conv_block(x_dim,y_dim),
            conv_block(y_dim,z_dim),
            Flatten())
    return PrototypicalNetworks(encoder)

class PrototypicalNetworks(nn.Module):
    def __init__(self,encoder):
        """
        Args:
            encoder : CNN encoding the images in sample
            n_way (int): number of classes in a classification task
            n_support (int): number of labeled examples per class in the support set
            n_query (int): number of labeled examples per class in the query set
        """
        super(PrototypicalNetworks, self).__init__()
        self.encoder = encoder

    def set_forward_loss(self, sample):
        """
        Computes loss, accuracy and output for classification task
        Args:
            sample (torch.Tensor): shape (n_way, n_support+n_query, (dim)) 
        Returns:
            torch.Tensor: shape(2), loss, accuracy and y_hat
        """
        sample_images = sample['images'].cuda()
        n_way = sample['n_way']
        n_support = sample['n_support']
        n_query = sample['n_query']

        x_support = sample_images[:, :n_support]
        x_query = sample_images[:, n_support:]

        #target indices are 0 ... n_way-1
        target_inds = torch.arange(0, n_way).view(n_way, 1, 1).expand(n_way, n_query, 1).long()
        target_inds = Variable(target_inds, requires_grad=False)
        target_inds = target_inds.cuda()

        #Gettig images in V stack
        z = sample_images
        z_dim=z.size(-1)
        z_proto = z[:n_way*n_support].view(n_way, n_support, z_dim).mean(1)
        z_query = z[n_way*n_support:]

        #compute distances
        dists = euclidean_dist(z_query, z_proto)

        #compute probabilities
        log_p_y = F.log_softmax(-dists, dim=1).view(n_way, n_query, -1)

        loss_val = -log_p_y.gather(2, target_inds).squeeze().view(-1).mean()
        _, y_hat = log_p_y.max(2)
        acc_val = torch.eq(y_hat, target_inds.squeeze()).float().mean()

        return loss_val, {
            'loss': loss_val.item(),
            'acc': acc_val.item(),
            'y_hat': y_hat}
                                                 


import torch
from torch import nn

__all__ = ['ConvBlock']

class ConvBlock(nn.Module):
    """
    Applies 'num_layers', 'stride_conv'(3x3) convolutions each followed by ReLU then downsamples via 'max_pool_stride'(2x2) max pool.
    """

    def __init__(self, num_layers=1, in_channels=4, out_channels=128,conv_kernel_size=3,stride_conv=3,padding=1,max_pool_kernal=2,max_pool_stride=2):
       super().__init__()
       self.convs = nn.ModuleList(
           [nn.Sequential( 
           nn.Conv2d(in_channels if i==0 else out_channels, out_channels, kernel_size=conv_kernel_size, stride=stride_conv,padding=padding),
           nn.ReLU()) 
           for i in range(num_layers)]
       )
       self.downsample = nn.MaxPool2d(kernel_size=max_pool_kernal, stride=max_pool_stride)
      
    def forward(self, x):
       for conv in self.convs:
           x = conv(x)
       x = self.downsample(x)
       return x
      


# Data load for proto net
# Path: src/metalearning/proto_dataloader.py

from mb_pytorch.dataloader.loader import data_fetcher
from mb_pytorch.utils.yaml_reader import YamlReader
from mb_pandas.src.dfload import load_any_df

__all__ = ['ProtoDataLoader']

class ProtoDataLoader(data_fetcher):
    """
    Data loader for proto net
    """
    def __init__(self, yaml, logger=None) -> None:
        super().__init__(yaml, logger)
        self.yaml = yaml
        self.logger = logger
        self._data = None
        self.embeddings = None
        self.data_dict = self.load_data_params()
        self.transforms_final = self.get_transforms()

    def load_embeddings(self):
        """
        load embeddings from yaml file
        Input:
            yaml: yaml file path (CSV or parquet)
        Output:
            embeddings: embeddings from yaml file as (CSV or parquet)
        """
        file = load_any_df(self.data_dict['data_train']['path'])
        self.embeddings = file['embeddings']
        return self.embeddings




#function: list of unet models
from torch.nn import functional as F
import torch
import torch.nn as nn
import torchvision
import torchvision.models as models
from .blocks import attention_block, conv_with_relu, rnn,conv_block


__all__ = ['Unet','Unet_attention']

class Unet_old(nn.Module):
    """
    Basic Unet model
    """
    def __init__(self,**kwargs):
        super(Unet_old,self).__init__()
        self._data=kwargs
        if 'conv_depth' in self._data:
            self.conv_depth=self._data['conv_depth']
        else:
            self.conv_depth=2
        if 'bottleneck_conv_depth' in self._data:
            self.bottleneck_conv_depth=self._data['bottleneck_conv_depth']
        else:
            self.bottleneck_conv_depth=self.conv_depth
        if 'unet_depth' in self._data:
            self.unet_depth=self._data['unet_depth']
        else:
            self.unet_depth=3
        if 'n_channels' in self._data:
            self.n_channels=self._data['n_channels']
        else:
            self.n_channels=3
        if 'n_classes' in self._data:
            self.n_classes=self._data['n_classes']
        else:
            self.n_classes=1
        if 'n_filters' in self._data:
            self.n_filters=self._data['n_filters']
        else:
            self.n_filters=64
        if 'linear_layers' in self._data:
            self.linear_layers=self._data['linear_layers']
        else:
            self.linear_layers=0

        self.in_channels = [self.n_channels] + [self.n_filters*(2**i) for i in range(self.unet_depth)]
        self.out_channels = self.in_channels[::-1][:-1] + [self.n_classes]
        
        #self.unet_conv = nn.Sequential()
        
        for i in range(self.unet_depth):
            temp_conv = conv_block.ConvBlock(in_channels=self.in_channels[i],out_channels=self.in_channels[i+1],
                                             depth=self.conv_depth,**kwargs)
            #self.unet_conv.add_module(f"unet_conv_{i}",temp_conv)
            self.add_module(f"unet_conv_{i}",temp_conv)
        

        #self.bottle_neck = nn.Sequential(conv_block.ConvBlock(in_channels=self.out_channels[0],out_channels=self.out_channels[0],
        #                                                      depth=self.bottleneck_conv_depth,sample_type='bottleneck',**kwargs))
        self.bottle_neck = conv_block.ConvBlock(in_channels=self.out_channels[0],out_channels=self.out_channels[0],
                                                depth=self.bottleneck_conv_depth,sample_type='bottleneck',**kwargs)

        #self.unet_deconv = nn.Sequential()
        for i in range(self.unet_depth-1):
            temp_deconv = conv_block.ConvBlock(in_channels=self.out_channels[i],out_channels=self.out_channels[i+1],
                                               depth=self.conv_depth,sample_type='up',**kwargs)
            #self.unet_deconv.add_module(f"unet_deconv_{i}",temp_deconv)
            self.add_module(f"unet_deconv_{i}",temp_deconv)

        #self.final = nn.Sequential()
        #self.final.add_module(f'final_conv',nn.Conv2d(self.in_channels[1],self.n_classes,kernel_size=1))
        #self.final.add_module(f'up_sample',nn.Upsample(scale_factor=2,mode='bilinear',align_corners=True))
        self.add_module(f'final_conv',nn.Conv2d(self.in_channels[1],self.n_classes,kernel_size=1))
        self.add_module(f'up_sample',nn.Upsample(scale_factor=2,mode='bilinear',align_corners=True))



        if self.linear_layers > 0:
            #self.final.add_module(f"unet_linear",nn.Linear(self.n_classes,self.n_classes))
            self.add_module(f"unet_linear",nn.Linear(self.n_classes,self.n_classes))

        if self.n_classes == 1:
            #self.final.add_module(f'final_activation',nn.Sigmoid())
            self.add_module(f'final_activation',nn.Sigmoid())
        else:
            #self.final.add_module(f'final_activation',nn.Softmax(dim=1))
            self.add_module(f'final_activation',nn.Softmax(dim=1))

    def forward(self,x):

        #x = self.unet_conv(x)
        x = self.unet_conv_0(x)
        x1 = self.unet_conv_1(x)
        x2 = self.unet_conv_2(x1)
        x3 = self.bottle_neck(x2)
        x4 = self.unet_deconv_0(x3+x2)
        x5 = self.unet_deconv_1(x4+x1)
        x6 = self.final_conv(x5+x)
        final = self.final_activation(self.up_sample(x6))
        return final


class Unet(nn.Module):
    """
    Basic Unet model
    """
    def __init__(self, conv_depth=2, bottleneck_conv_depth=None, unet_depth=3, n_channels=3, n_classes=1, n_filters=64, linear_layers=0, **kwargs):
        super(Unet, self).__init__()
        if 'conv_depth' in kwargs:
            self.conv_depth = kwargs['conv_depth']
        else:
            self.conv_depth = conv_depth
        if 'bottleneck_conv_depth' in kwargs:
            self.bottleneck_conv_depth = kwargs['bottleneck_conv_depth']
        else:
            self.bottleneck_conv_depth = bottleneck_conv_depth if bottleneck_conv_depth is not None else conv_depth
        if 'unet_depth' in kwargs:
            self.unet_depth = kwargs['unet_depth']
        else:
            self.unet_depth = unet_depth
        if 'n_channels' in kwargs:
            self.n_channels = kwargs['n_channels']
        else:
            self.n_channels = n_channels
        if 'n_classes' in kwargs:
            self.n_classes = kwargs['n_classes']
        else:
            self.n_classes = n_classes
        if 'n_filters' in kwargs:
            self.n_filters = kwargs['n_filters']
        else:
            self.n_filters = n_filters
        if 'linear_layers' in kwargs:
            self.linear_layers = kwargs['linear_layers']
        else:
            self.linear_layers = linear_layers

        self.in_channels = [self.n_channels] + [self.n_filters*(2**i) for i in range(self.unet_depth)]
        self.out_channels = self.in_channels[::-1][:-1] + [self.n_classes]

        self.unet_conv = nn.ModuleList([conv_block.ConvBlock(in_channels=self.in_channels[i], out_channels=self.in_channels[i+1], **kwargs) for i in range(self.unet_depth)])

        self.bottle_neck = conv_block.ConvBlock(in_channels=self.out_channels[0],out_channels=self.out_channels[0], **kwargs)

        self.unet_deconv = nn.ModuleList([conv_block.ConvBlock(in_channels=self.out_channels[i], out_channels=self.out_channels[i+1], sample_type='up', **kwargs) for i in range(self.unet_depth-1)])

        self.final_conv = nn.Conv2d(self.in_channels[1], self.n_classes, kernel_size=1)
        self.up_sample = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        if self.linear_layers > 0:
            self.unet_linear = nn.Linear(self.n_classes,self.n_classes)

        if self.n_classes == 1:
            self.final_activation = nn.Sigmoid()
        else:
            self.final_activation = nn.Softmax(dim=1)

    def forward(self, x):
        temp_out = []
        for i in range(self.unet_depth):
            x = self.unet_conv[i](x)
            temp_out.append(x)
            x = temp_out[-1]

        x = self.bottle_neck(x)

        for i in range(self.unet_depth-1):
            x = self.unet_deconv[i](x + temp_out[-(i+1)])

        x = self.final_conv(x + temp_out[0])
        x = self.up_sample(x)
        final = self.final_activation(x)

        if self.linear_layers > 0:
            final = final.view(final.size(0),-1)
            final = self.unet_linear(final)
            
        return final
    

class Unet_attention(nn.Module):
    """
    Unet with attnetion module
    """
    def __init__(self, conv_depth=2, bottleneck_conv_depth=None, unet_depth=3, n_channels=3, n_classes=1, n_filters=64, linear_layers=0, **kwargs):
        super(Unet_attention, self).__init__()
        if 'conv_depth' in kwargs:
            self.conv_depth = kwargs['conv_depth']
        else:
            self.conv_depth = conv_depth
        if 'bottleneck_conv_depth' in kwargs:
            self.bottleneck_conv_depth = kwargs['bottleneck_conv_depth']
        else:
            self.bottleneck_conv_depth = bottleneck_conv_depth or conv_depth
        if 'unet_depth' in kwargs:
            self.unet_depth = kwargs['unet_depth']
        else:
            self.unet_depth = unet_depth
        if 'n_channels' in kwargs:
            self.n_channels = kwargs['n_channels']
        else:
            self.n_channels = n_channels
        if 'n_classes' in kwargs:
            self.n_classes = kwargs['n_classes']
        else:
            self.n_classes = n_classes
        if 'n_filters' in kwargs:
            self.n_filters = kwargs['n_filters']
        else:
            self.n_filters = n_filters
        if 'linear_layers' in kwargs:
            self.linear_layers = kwargs['linear_layers']
        else:
            self.linear_layers = linear_layers

        self.in_channels = [self.n_channels] + [self.n_filters*(2**i) for i in range(self.unet_depth)]
        self.out_channels = self.in_channels[::-1][:-1] + [self.n_classes]

        self.unet_conv = nn.ModuleList([conv_block.ConvBlock(in_channels=self.in_channels[i], out_channels=self.in_channels[i+1], depth=self.conv_depth, **kwargs) for i in range(self.unet_depth)])

        self.bottle_neck = conv_block.ConvBlock(in_channels=self.out_channels[0],out_channels=self.out_channels[0], depth=self.bottleneck_conv_depth, sample_type='bottleneck',**kwargs)

        self.unet_deconv = nn.ModuleList([conv_block.ConvBlock(in_channels=self.out_channels[i], out_channels=self.out_channels[i+1], depth=self.conv_depth, sample_type='up', **kwargs) for i in range(self.unet_depth-1)])
    
        self.attention = nn.ModuleList([attention_block.Attention(x=self.out_channels[i],g=self.out_channels[i]) for i in range(self.unet_depth-1)])

        self.final_conv = nn.Conv2d(self.in_channels[1], self.n_classes, kernel_size=1)
        self.up_sample = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        if self.linear_layers > 0:
            self.unet_linear = nn.Linear(self.n_classes,self.n_classes)

        if self.n_classes == 1:
            self.final_activation = nn.Sigmoid()
        else:
            self.final_activation = nn.Softmax(dim=1)

    def forward(self, x):
        temp_out = []
        for i in range(self.unet_depth):
            x = self.unet_conv[i](x)
            temp_out.append(x)
            x = temp_out[-1]

        x = self.bottle_neck(x)

        for i in range(self.unet_depth-1):
            x = self.attention[i](x,temp_out[-(i+1)])
            x = self.unet_deconv[i](x + temp_out[-(i+1)])

        x = self.final_conv(x + temp_out[0])
        x = self.up_sample(x)
        final = self.final_activation(x)

        if self.linear_layers > 0:
            final = final.view(final.size(0),-1)
            final = self.unet_linear(final)
            
        return final